@@ -1,10 +1,8 @@
-import dataclasses
-import enum
 import math
 import multiprocessing
 import threading
 import tkinter as tk
-from typing import Any
+from typing import Any, Optional
 
 try:
     from ico import load_ico
@@ -41,11 +39,6 @@ def get_dpi() -> float:
     return current_dpi
 
 
-class State(enum.Enum):
-    COLLAPSED = 0
-    EXPANDED = 1
-
-
 class TreeNode:
     font_size: int = 10
     # The values below will be multiplied by font_size
@@ -57,11 +50,11 @@ class TreeNode:
     line_style: dict[str, Any] = {'fill': 'gray60'}  # , 'dash': (1, 1)}
     label_style: dict[str, dict[str, Any]] = {
         'normal': {
-            'foreground': '#000000',
-            'background': '#ffffff',
+            'foreground': 'black',
+            'background': 'gray97',
         },
         'selected': {
-            'foreground': '#000000',
+            'foreground': 'black',
             'background': 'gray60',
         },
     }
@@ -82,16 +75,17 @@ class TreeNode:
         cls.text_pad = round(size*cls.text_pad_factor)
         cls.icon_size = size*cls.icon_size_factor
 
-    def __init__(self, root, parent, object_tree):
+    def __init__(self, root: 'ViewTreeWindow', parent: Optional['TreeNode'],
+                 tree: 'PicklableTree'):
         self.font: tuple[str, int] = ('Consolas', self.font_size)
         # Cache of PhotoImage instances for icons
-        self.icon_images: dict[str, tk.PhotoImage] = {}
+        self.icons: dict[str, tk.PhotoImage] = {}
         self.root: ViewTreeWindow = root
         self.canvas: tk.Canvas = root.canvas
         self.statusbar: tk.Label = root.statusbar
         self.parent: TreeNode | None = parent
-        self.object_tree: PicklableObjectTree = object_tree
-        self.state: State = State.EXPANDED
+        self.tree: PicklableTree = tree
+        self.is_expanded: bool = True
         self.is_visible: bool = True
         self.is_selected: bool = False
         self.children: list[TreeNode] = []
@@ -102,22 +96,19 @@ class TreeNode:
 
         self.root.n_lines += 1
         self.lines_maxed: bool = False
-        if not self.object_tree.too_deep:
-            self.root.all_nodes.append(self)
-            if self.object_tree.overflowed:
-                self.root.n_lines += 1
-        for branch in object_tree.visible_branches:
-            if self.root.n_lines >= self.object_tree.max_lines:
+        self.root.all_nodes.append(self)
+        if self.tree.overflowed or self.tree.maxed_depth:
+            self.root.n_lines += 1
+        for branch in tree:
+            if self.root.n_lines >= self.tree.max_lines:
                 self.lines_maxed = True
                 break
             self.children.append(TreeNode(self.root, self, branch))
-            if branch.too_deep:
-                break
 
-    def update(self):
+    def refresh(self):
         """Refresh the whole canvas"""
         if self.parent:
-            self.parent.update()
+            self.parent.refresh()
         else:
             self.canvas['cursor'] = 'watch'
             self.canvas.update()
@@ -136,7 +127,7 @@ class TreeNode:
 
     def update_visibility(self):
         """Update the .is_visible flags"""
-        if self.is_visible and self.state == State.EXPANDED:
+        if self.is_visible and self.is_expanded:
             for child in self.children:
                 child.is_visible = True
                 child.update_visibility()
@@ -147,64 +138,74 @@ class TreeNode:
 
     def draw(self, x=0, y=0) -> int:
         """Draw a node and recursively call itself for each child node.
-        Return the next available vertical position for drawing."""
+        Return the next available vertical position for drawing.
+        """
         dy = self.row_height//2
-        icon_width = self.draw_icon(self, x + dy, y + dy)
+        icon_width = self.draw_icon(x + dy, y + dy)
         self.x = x + (self.row_height + icon_width)//2 + self.text_pad
         self.y = y
-        if self.object_tree.too_deep:
-            # Max depth reached -- draw ellipsis
-            self.draw_overflow(self.x, y)
-            return y + self.row_height
         self.draw_text()
         x2 = x + self.indent_width
         y2 = y + self.row_height
 
-        if not (self.children or self.lines_maxed):
+        if (not self.children and not self.lines_maxed
+                and not self.tree.maxed_depth):
             return y2
-        if self.state == State.COLLAPSED:
+        if not self.is_expanded:
             return y2
 
         line_x = x + dy
         line_y = y2 + dy
         last_y = line_y
+
+        if (self.tree.is_expandable and self.tree.maxed_depth
+                and not self.tree.overflowed):
+            # If max depth reached, draw ellipsis
+            last_y = line_y
+            # Horizontal line
+            self.draw_line(line_x, line_y, self.indent_width, 0)
+            self.draw_overflow(x2 + dy + self.text_pad, y2)
+            y2 += self.row_height
+            line_y = y2 + dy
+
         for child in self.children:
             last_y = line_y
             # Horizontal line
-            self.canvas.create_line(line_x, line_y,
-                                    line_x + self.indent_width, line_y,
-                                    **self.line_style)
+            self.draw_line(line_x, line_y, self.indent_width, 0)
             y2 = child.draw(x2, y2)
             line_y = y2 + dy
-        # noinspection PyUnboundLocalVariable
-        if (self.lines_maxed or (self.object_tree.overflowed
-                                 and not child.object_tree.too_deep)):
-            # Max lines or max branches reached -- draw ellipsis
+
+        if self.lines_maxed or self.tree.overflowed:
+            # If max lines or max branches reached, draw ellipsis
             self.draw_overflow(line_x - self.text_pad, y2)
             last_y = y2 + self.row_height//4  # Small additional line
             y2 += self.row_height
 
         # Vertical line
-        _id = self.canvas.create_line(line_x, y + dy, line_x, last_y,
-                                      **self.line_style)
+        _id = self.draw_line(line_x, y + dy, 0, last_y - y - dy)
         self.canvas.tag_lower(_id)  # Display under the icons
 
         return y2
 
-    def draw_icon(self, node, x, y) -> int:
+    def draw_line(self, x: int, y: int, dx: int, dy: int) -> int:
+        return self.canvas.create_line(x, y, x + dx, y + dy, **self.line_style)
+
+    def draw_icon(self, x: int, y: int) -> int:
         """Draw the plus or the minus icon if it is expandable. Return
-        the icon width."""
-        if not node.object_tree.expandable:
+        the icon width.
+        """
+        if not self.tree.is_expandable:
             return 0
-        if node.state == State.EXPANDED:
+        if self.is_expanded:
             icon_name = 'minus'
-            callback = node.collapse
+            callback = self.collapse
         else:
             icon_name = 'plus'
-            callback = node.expand
-        image = self.get_icon_image(icon_name)
-        _id = self.canvas.create_image(x, y, image=image)
+            callback = self.expand
+        icon = self.get_icon(icon_name)
+        _id = self.canvas.create_image(x, y, image=icon)
         self.canvas.tag_bind(_id, '<1>', callback)
+        self.canvas.tag_bind(_id, '<3>', self.toggle_children)
         self.canvas.tag_bind(_id, '<Double-1>', callback)
         self.canvas.tag_bind(_id, '<Enter>', self.update_cursor)
         self.canvas.tag_bind(_id, '<Leave>', self.update_cursor)
@@ -213,13 +214,13 @@ class TreeNode:
         y = self.root.winfo_pointery()
         if x1 <= x <= x2 and y1 <= y <= y2:
             self.canvas['cursor'] = 'hand2'
-        return image.width()
+        return icon.width()
 
     def draw_text(self):
         """Draw the key and the value type of the node"""
         text_x = self.x
         text_y = self.y
-        text = self.object_tree.node_str or ''
+        text = self.tree.node_str or ''
         if self.label is None:
             self.label = tk.Label(self.canvas, text=text, bd=0,
                                   padx=self.text_pad, pady=self.text_pad,
@@ -243,7 +244,8 @@ class TreeNode:
             font=self.font, fill=self.label_style['normal']['foreground']
         )
 
-    def select(self, _=None, update_yview=True, update_last_selected=True):
+    def select(self, _event=None,
+               update_yview=True, update_last_selected=True):
         # Selection can switch directly or as a result of expanding or
         # collapsing node. If a selected node is collapsed, it remembers
         # the last selected node so that it can be highlighted after it
@@ -256,7 +258,7 @@ class TreeNode:
         self.is_selected = True
         if update_last_selected:
             self.root.last_selected = self
-            path = self.object_tree.path
+            path = self.tree.path
             text = f'Path: {path}' if path else ''
             self.root.statusbar.config(text=text)
         self.root.selected = self
@@ -264,7 +266,7 @@ class TreeNode:
         if update_yview:
             self.update_yview()
 
-    def deselect(self, _=None):
+    def deselect(self, _event=None):
         if not self.is_selected:
             return
         self.is_selected = False
@@ -277,30 +279,77 @@ class TreeNode:
             return
         self.root.selected.deselect()
 
-    def expand(self, _=None):
-        if not self.object_tree.expandable:
+    def expand(self, _event=None):
+        if not self.tree.is_expandable:
             self.select()
             return
-        if self.state == State.COLLAPSED:
-            self.state = State.EXPANDED
-            self.update_visibility()
-            self.update()
-            if child := self.has_last_selected():
-                child.select(update_yview=False, update_last_selected=False)
-            self.update_yview(from_expand=True)
-
-    def collapse(self, _=None):
-        if not self.object_tree.expandable:
+        if self.is_expanded:
             return
-        if self.state == State.EXPANDED:
-            self.state = State.COLLAPSED
+        self.is_expanded = True
+        self.update_visibility()
+        self.refresh()
+        if child := self.has_last_selected():
+            child.select(update_yview=False, update_last_selected=False)
+        self.update_yview(from_expand=True)
+
+    def collapse(self, _event=None):
+        if not self.tree.is_expandable:
+            return
+        if not self.is_expanded:
+            return
+        self.is_expanded = False
+        self.update_visibility()
+        self.refresh()
+        # Remember the last selected if it is contained in the
+        # collapsed nodes
+        if (self.root.last_selected is not None
+                and self.root.last_selected.is_descendant(self)):
+            self.select(update_last_selected=False)
+
+    def toggle_children(self, _event=None):
+        """Toggle the expansion state of all child nodes. Use the
+        majority state as the reference and apply the inverse state to
+        every child node.
+        """
+        if not self.tree.is_expandable:
+            self.select()
+            return
+        if not self.children:
+            return
+        collapse: bool = bool(round(
+            sum(
+                child.is_expanded for child in self.children
+            )/len(self.children)
+        ))
+        if not self.is_expanded:
+            self.is_expanded = True
             self.update_visibility()
-            # Remember the last selected if it is contained in the
-            # collapsed nodes
-            if (self.root.last_selected is not None and
-                    self.root.last_selected.is_descendant(self)):
-                self.select(update_last_selected=False)
-            self.update()
+            collapse = True
+        if collapse:
+            next_selected: TreeNode | None = None
+            for child in self.children:
+                child.is_expanded = False
+                child.update_visibility()
+                # Remember the last selected if it is contained in the
+                # collapsed nodes
+                if (next_selected is None
+                        and self.root.last_selected is not None
+                        and self.root.last_selected.is_descendant(child)):
+                    next_selected = child
+            self.refresh()
+            if next_selected is not None:
+                next_selected.select(update_last_selected=False)
+        else:  # expand
+            last_selected: TreeNode | None = None
+            for child in self.children:
+                child.is_expanded = True
+                child.update_visibility()
+                last_selected = last_selected or child.has_last_selected()
+            self.refresh()
+            if last_selected is not None:
+                last_selected.select(update_yview=False,
+                                     update_last_selected=False)
+            self.update_yview(from_expand=True)
 
     def update_yview(self, from_expand=False):
         # If it comes from an expanding node, move the canvas up so that
@@ -354,33 +403,33 @@ class TreeNode:
 
     def visible_children_bottom(self):
         """Bottom y coordinate of visible children"""
-        if self.children and self.state == State.EXPANDED:
+        if self.children and self.is_expanded:
             bottom = self.children[-1].visible_children_bottom()
-            if self.object_tree.overflowed:
+            if self.tree.overflowed:
                 return bottom + self.row_height
             return bottom
-        elif self.object_tree.overflowed:
+        elif self.tree.overflowed:
             return self.y + 2*self.row_height
         return self.y + self.row_height
 
-    def get_icon_image(self, file_name: str) -> tk.PhotoImage:
+    def get_icon(self, file_name: str) -> tk.PhotoImage:
         """Load the plus/minus icons"""
-        if file_name in self.icon_images:
-            return self.icon_images[file_name]
+        if file_name in self.icons:
+            return self.icons[file_name]
         try:
             icon_data = self.icon_files[file_name]
         except KeyError:
             icon_data = load_ico(file_name, self.icon_size)
             self.icon_files[file_name] = icon_data
         image = tk.PhotoImage(master=self.canvas, data=icon_data)
-        self.icon_images[file_name] = image
+        self.icons[file_name] = image
         return image
 
-    def copy(self, _=None):
+    def copy(self, _event=None):
         """Copy to clipboard. Called on Ctrl+C or double-click"""
         if not self.is_selected:
             raise Exception('Copy called without being selected')
-        if not (path := self.object_tree.path):
+        if not (path := self.tree.path):
             return 'break'
         if _PYPERCLIP_LOADED:
             # Preferred method
@@ -403,8 +452,8 @@ class TreeNode:
 
 class ScrollbarFrame(tk.Frame):
     """A frame with vertical and horizontal scrollbars. The horizontal
-    one automatically hides when it is disabled."""
-
+    one automatically hides when it is disabled.
+    """
     def __init__(self, master, *, row_height=5, mouse_wheel_rows=3, **opts):
         self.row_height = row_height
         self.mouse_wheel_rows = mouse_wheel_rows
@@ -441,7 +490,7 @@ class ScrollbarFrame(tk.Frame):
         self.on_resize()
         self.canvas.focus_set()
 
-    def on_resize(self, _=None):
+    def on_resize(self, _event=None):
         """Reset the scroll region to encompass the inner frame"""
         _, _, x, y = self.canvas.bbox(tk.ALL)
         w = self.canvas.winfo_width()
@@ -528,15 +577,16 @@ class ScrollbarFrame(tk.Frame):
 class ViewTreeWindow(tk.Tk):
     width = 540
     height = 720
+    background_color = 'gray97'
 
-    def __init__(self, object_tree):
+    def __init__(self, tree):
         super().__init__()
         self.title('Tree View')
         self.geometry(f'{self.width}x{self.height}')
 
         TreeNode.update_dpi()
         self.sf = ScrollbarFrame(self, row_height=TreeNode.row_height,
-                                 bg='white', borderwidth=0,
+                                 bg=self.background_color, borderwidth=0,
                                  highlightthickness=0, takefocus=1)
         self.canvas = self.sf.canvas
         self.statusbar = tk.Label(self, text='', bd=1,
@@ -549,9 +599,9 @@ class ViewTreeWindow(tk.Tk):
         self.last_selected: TreeNode | None = None
         self.all_nodes: list[TreeNode] = []
         self.n_lines: int = 0
-        self.object_tree: PicklableObjectTree = object_tree
-        self.node: TreeNode = TreeNode(self, None, object_tree)
-        self.node.update()
+        self.tree: PicklableTree = tree
+        self.node: TreeNode = TreeNode(self, None, tree)
+        self.node.refresh()
         self.node.expand()
         self.node.select()
 
@@ -560,6 +610,7 @@ class ViewTreeWindow(tk.Tk):
         self.bind('<Key-Down>', self.move_down)
         self.bind('<Key-Left>', self.collapse)
         self.bind('<Key-Right>', self.expand)
+        self.bind('<space>', self.toggle_children)
 
         try:
             images_data = load_ico('icon.ico')
@@ -573,12 +624,12 @@ class ViewTreeWindow(tk.Tk):
     def on_resize(self):
         self.sf.on_resize()
 
-    def copy(self, _=None):
+    def copy(self, _event=None):
         if self.selected is None:
             return 'break'
         self.selected.copy()
 
-    def move_up(self, _=None):
+    def move_up(self, _event=None):
         if self.selected is None:
             return 'break'
         index = max(self.selected.index, 1)
@@ -587,7 +638,7 @@ class ViewTreeWindow(tk.Tk):
                 node.select()
                 return
 
-    def move_down(self, _=None):
+    def move_down(self, _event=None):
         if self.selected is None:
             return 'break'
         index = self.selected.index
@@ -596,7 +647,7 @@ class ViewTreeWindow(tk.Tk):
                 node.select()
                 break
 
-    def collapse(self, _=None):
+    def collapse(self, _event=None):
         if self.selected is None:
             return 'break'
         if self.selected.is_visible:
@@ -604,7 +655,7 @@ class ViewTreeWindow(tk.Tk):
         else:
             self.move_up()
 
-    def expand(self, _=None):
+    def expand(self, _event=None):
         if self.selected is None:
             return 'break'
         if self.selected.is_visible:
@@ -612,30 +663,31 @@ class ViewTreeWindow(tk.Tk):
         else:
             self.move_down()
 
-
-@dataclasses.dataclass
-class PicklableObjectTree:
-    visible_branches: tuple['PicklableObjectTree', ...]
-    node_str: str
-    path: str
-    expandable: bool
-    overflowed: bool
-    too_deep: bool
-    max_lines: float
-
-    def __init__(self, object_tree):
-        self.visible_branches = tuple(map(PicklableObjectTree,
-                                          object_tree.visible_branches))
-        self.node_str: str = object_tree.node_text
-        self.path: str = object_tree.path
-        self.expandable: bool = object_tree.expandable
-        self.overflowed: bool = object_tree.overflowed
-        self.too_deep: bool = object_tree.too_deep
-        self.max_lines: int = object_tree.config.max_lines
+    def toggle_children(self, _event=None):
+        if self.selected is None or not self.selected.is_visible:
+            return 'break'
+        self.selected.toggle_children()
 
 
-def tree_window_loop(object_tree: PicklableObjectTree):
+class PicklableTree(tuple):
+    """Transform Tree into a simpler pickable object"""
 
+    def __new__(cls, tree):
+        return super().__new__(cls, map(PicklableTree, tree))
+
+    def __init__(self, tree):
+        self.node_str: str = tree.node_text
+        self.path: str = tree.path
+        self.is_expandable: bool = tree.is_expandable
+        self.overflowed: bool = tree.overflowed
+        self.maxed_depth: bool = tree.maxed_depth
+        self.max_lines: float = tree.config.max_lines
+
+
+def tree_window_loop(tree: PicklableTree):
+    """Main loop for the GUI. Can be used concurrently, as a separate
+    thread, or as a separate process.
+    """
     # Make text look sharper for displays with scaling different
     # from 100%. Only works for Windows.
     # WARNING: This is permanently applied to the whole process, not
@@ -643,17 +695,20 @@ def tree_window_loop(object_tree: PicklableObjectTree):
     if SET_DPI_AWARENESS and windll is not None:
         windll.shcore.SetProcessDpiAwareness(1)
 
-    window = ViewTreeWindow(object_tree)
+    window = ViewTreeWindow(tree)
     window.mainloop()
 
 
-def tree_viewer(object_tree, *, spawn_thread=True, spawn_process=False):
-    object_tree = PicklableObjectTree(object_tree)
+def tree_viewer(tree, *, spawn_thread=True, spawn_process=False):
+    """GUI spawner. Convert Tree to a simpler pickable object and
+    optionally spawn a new thread or process.
+    """
+    tree = PicklableTree(tree)
     if spawn_process:
         multiprocessing.Process(target=tree_window_loop,
-                                args=(object_tree,)).start()
+                                args=(tree,)).start()
     elif spawn_thread:
-        threading.Thread(target=tree_window_loop, args=(object_tree,),
+        threading.Thread(target=tree_window_loop, args=(tree,),
                          daemon=True).start()
     else:
-        tree_window_loop(object_tree)
+        tree_window_loop(tree)
